#!tclsh
# $Header: migrate_to_mx.tcl,v 1.10  $
# Migrate 1200XL and 3200DX designs to 42MX


# Usage: designer script:$argv0+<original.adb>+<newname.adb> logfile:<newname.log>

proc usage { argsExpected } {
    global argv0

    if { $argsExpected == 2 } {
	puts stderr "Usage: designer script:$argv0+<original.adb>+<newname.adb> logfile:<newname.log>\n"
    } else {
	puts stderr "Usage: $argv0 <newname.adb>\n"
    }
}


# Open ADB, if not already open (if invoked within Designer)
# Check if basename of newAdb already exists in adb, adl, pin, crt, loc or log
# Ensure that newAdb file is writeable
# Return the new ADB filename

proc openAdb {} {
    global argc
    global argv0
    global argv
    set argsExpected 1

    if { [catch "get_defvar FAM" fam] || ($fam == {}) } {
	incr argsExpected
    }

    if { $argc != $argsExpected } {
	usage $argsExpected
	return {}
    }

    if { $argsExpected == 2 } {
	foreach {origAdb newAdb} $argv break
    } else {
	foreach {newAdb} $argv break
    }

    if { [file extension $newAdb] != ".adb" } {
	usage $argsExpected
	puts stderr " Target file $newAdb does not have .adb extension."
	return {}
    }
    set newBase [file rootname $newAdb]
    foreach ext {"adb" "adl" "pin" "crt" "loc" "log"} {
	if { [file readable "$newBase.$ext"] } {
	    usage $argsExpected
	    puts stderr " File $newBase.$ext exists."
	    puts stderr " Please provide a unique base name for target file <newname.adb>."
	    return {}
	}
    }

    if { $argsExpected == 2 } {
	if { ![file readable $origAdb] } {
	    usage $argsExpected
	    puts stderr " File $origAdb not found."
	    return {}
	}
	if { [catch "open_design $origAdb" msg] } {
	    usage $argsExpected
	    puts stderr "$msg Cannot open $origAdb file."
	    return {}
	}
    }
    if { [catch "save_design $newAdb" msg] } {
	usage $argsExpected
	puts stderr "$msg Cannot write target file $newAdb."
	return {}
    }
    set argv0 [file tail $argv0]

    return [list $newAdb $argsExpected]
}


# Cleanup files with basename of newAdb and extension in exts

proc cleanup { newAdb exts } {
    set newBase [file rootname $newAdb]

    foreach ext $exts {
	catch "file delete -force -- $newBase.$ext"
    }
}


# Is the design state in placed or compiled state

proc isPlaced {} {
    global argv0

    if { [catch "is_design_state_complete LAYOUT" placed] } {
	set placed 0
    }
    if { $placed } {
	return 1
    }
    if { [catch "is_design_state_complete COMPILE" compiled] || !$compiled } {
	puts stderr "$argv0: design is not compiled, aborting"
	return -1
    }
    return 0
}


# get_defvar with error message

proc defget { var } {
    global argv0

    if { [catch "get_defvar $var" val] || ($val == {}) } {
	puts stderr "$argv0: cannot retrieve $var setting, aborting"
	return {}
    }
    return $val
}


# Check if netlist has 3200DX specific macros

proc findWidedecRamQclock { adlFile } {
    global argv0
    set widedec 0
    set ram 0
    set qclock 0

    if { [catch "open $adlFile r" fp] } {
	puts stderr "$argv0: read $adlFile failed"
	return {}
    }
    while { [gets $fp line] >= 0 } {
	switch -regexp -- $line {
	    {^[Uu][Ss][Ee] +[Aa][Dd][Ll][Ii][Bb]:(|[Dd][Xx])[Rr][Aa][Mm]} {
		incr ram
	    }
	    {^[Uu][Ss][Ee] +[Aa][Dd][Ll][Ii][Bb]:[Dd][Xx]} {
		incr widedec
	    }
	    {^[Uu][Ss][Ee] +[Aa][Dd][Ll][Ii][Bb]:[Qq][Cc][Ll][Kk]} {
		incr qclock
	    }
	}
    }
    close $fp

    return [list $widedec $ram $qclock]
}


# Map 1200XL and 3200DX devices to the most appropriate 42MX die
# Also, return the new FAM
# Indicate if it belonged to 3200DX
# And, are the two devices P&R compatible

proc mapDie { adlFile } {
    global argv0
    set newFam "42MX"
    set isPlaceCompatible 1
    set isDX 0
    set die [defget "DIE"]

    switch -regexp -- $die {
	{^$} {
	    return {}
	}
	{1[02][28]0[Rr][HhTt]} {
	    puts stderr "$argv0: Radiation device $die cannot be migrated, aborting\n"
	    return {}
	}
	{1225[Xx][Ll]} {
	    set newDie "${newFam}09"
	    set isPlaceCompatible 0
	}
	{1240[Xx][Ll]} {
	    set newDie "${newFam}09"
	}
	{1280[Xx][Ll]} {
	    set newDie "${newFam}16"
	}
	{3265} {
	    incr isDX
	    set isPlaceCompatible 0
	    set dxMacros [findWidedecRamQclock $adlFile]
	    if { $dxMacros == {} } {
		return {}
	    }
	    foreach {widedec ram qclock} $dxMacros break

	    if { !$widedec && !$ram && !$qclock } {
		set newDie "${newFam}16"
		set isDX 0
	    } elseif { !$ram && !$qclock } {
		set newDie "${newFam}24"
	    } else {
		set newDie "${newFam}36"
	    }
	}
	{32100} {
	    incr isDX
	    set isPlaceCompatible 0
	    set dxMacros [findWidedecRamQclock $adlFile]
	    if { $dxMacros == {} } {
		return {}
	    }
	    foreach {widedec ram qclock} $dxMacros break

	    if { !$ram && !$qclock } {
		set newDie "${newFam}24"
	    } else {
		set newDie "${newFam}36"
	    }
	}
	{32140} {
	    incr isDX
	    set newDie "${newFam}24"
	}
	{32200} {
	    incr isDX
	    set newDie "${newFam}36"
	}
	default {
	    puts stderr "$argv0: Device $die cannot be migrated, aborting\n"
	    return {}
	}
    }
    return [list $die $newFam $newDie $isDX $isPlaceCompatible]
}


# Check if an exact or a superset package is available in new device

proc mapPackage { die } {
    global argv0

    if { ([set pkg [defget "PACKAGE"]] == {}) ||
	 ([set allPkgs [defget "ALL_${die}_PACKAGES"]] == {}) } {
	return {}
    }
    regsub {([^0-9]*)[0-9]*.*} $pkg {\1} pkgType
    regsub {[^0-9]*([0-9]*).*} $pkg {\1} pkgPins

    switch -regexp -- $pkgType {
	{rdqfp} {
	    set pkgType [list $pkgType "qfp"]
	}
	{qfp} {
	    set pkgType [list $pkgType "cqfp"]
	}
    }
    set supersetPins 0

    foreach newPkg [split $allPkgs] {
	if { [string match -nocase $pkg $newPkg] } {
	    return [list $pkg $newPkg]
	}
	regsub {([^0-9]*)[0-9]*.*} $newPkg {\1} newPkgType
	regsub {[^0-9]*([0-9]*).*} $newPkg {\1} newPkgPins

	foreach type $pkgType {
	    if { [string match -nocase $type $newPkgType] &&
		 ($newPkgPins >= $pkgPins) &&
		 ($newPkgPins > $supersetPins) } {
		set supersetType $newPkgType
		set supersetPins $newPkgPins
	    }
	}
    }
    if { !$supersetPins } {
	puts stderr "$argv0: package $pkg is not available in $die"
	return {}
    }
    set newPkg "${supersetType}${supersetPins}"
    puts "WARNING: package $pkg is not available in $die, converting to $newPkg"
    return [list $pkg $newPkg]
}


# Check if original speed-grade exists in new device, otherwise find the default

proc mapSpeedGrade { fam die pkg } {

    set speed [defget "SPEED"]
    if { $speed == {} } {
	return {}
    }
    catch "get_defvar ALL_${die}_${pkg}_SPEEDS" allSpeeds
    if { $allSpeeds == {} } {
	catch "get_defvar ALL_${die}_SPEEDS" allSpeeds
	if { $allSpeeds == {} } {
	    catch "get_defvar ALL_${fam}_SPEEDS" allSpeeds
	}
    }
    foreach newSpeed [split $allSpeeds] {
	if { [string match -nocase $speed $newSpeed] } {
	    return [list $speed $newSpeed]
	}
    }
    catch "get_defvar ${die}_${pkg}_SPEED" newSpeed
    if { $newSpeed == {} } {
	catch "get_defvar ${die}_SPEED" newSpeed
	if { $newSpeed == {} } {
	    catch "get_defvar ${fam}_SPEED" newSpeed
	    if { $newSpeed == {} } {
		set newSpeed "STD"
	    }
	}
    }
    puts "WARNING: Speed-grade $speed is not available in ${die}_${pkg}, converting to $newSpeed"
    return [list $speed $newSpeed]
}


# Get all Temperature and Voltage ranges for the new device

proc getAllRanges { fam die pkg speed voltage } {

    catch "get_defvar ALL_${die}_${pkg}_${speed}_${voltage}_RANGES" allRanges
    if { $allRanges == {} } {
	catch "get_defvar ALL_${die}_${pkg}_${speed}_RANGES" allRanges
	if { $allRanges == {} } {
	    catch "get_defvar ALL_${die}_${pkg}_RANGES" allRanges
	    if { $allRanges == {} } {
		catch "get_defvar ALL_${die}_RANGES" allRanges
		if { $allRanges == {} } {
		    catch "get_defvar ALL_${fam}_${speed}_${voltage}_RANGES" allRanges
		    if { $allRanges == {} } {
			catch "get_defvar ALL_${fam}_${voltage}_RANGES" allRanges
			if { $allRanges == {} } {
			    catch "get_defvar ALL_${fam}_${speed}_RANGES" allRanges
			    if { $allRanges == {} } {
				catch "get_defvar ALL_${fam}_RANGES" allRanges
				if { $allRanges == {} } {
				    set allRanges {COM IND}
				}
			    }
			}
		    }
		}
	    }
	}
    }
    return $allRanges
}


# Check if original Temperature or Voltage range exists in new device

proc tempVoltRangeExists { var allRanges } {
    global argv0

    set range [defget $var]
    if { $range == {} } {
	return {}
    }
    if { [llength [split $range]] > 1 } {
	return [list $range]
    }
    foreach newRange [split $allRanges] {
	if { [string match -nocase $range $newRange] } {
	    return $newRange
	}
    }
    puts stderr "$argv0: ${var}ange $range is not available, aborting"
    return {}
}


# Create the set_design command

proc setDesignCmd { newFam } {

    set design [defget "DESIGN"]
    if { $design == {} } {
	return {}
    }
    return [list set_design -name $design -family $newFam]
}


# Create the set_device command

proc setDeviceCmd { newFam newDie newPkg newSpeed isDX } {

    if { [set voltage [defget "VOLTAGE"]] == {} ||
	 [set allRanges [getAllRanges $newFam $newDie $newPkg $newSpeed $voltage]] == {} ||
	 [set tempr [tempVoltRangeExists "TEMPR" $allRanges]] == {} ||
	 [set voltr [tempVoltRangeExists "VOLTR" $allRanges]] == {} ||
	 [set probe [defget "RESTRICTPROBEPINS"]] == {} } {
	return {}
    }
    set cmd [list set_device -die $newDie	\
			     -package $newPkg	\
			     -speed $newSpeed	\
			     -voltage $voltage	\
			     -temprange $tempr	\
			     -voltrange $voltr	\
			     -probe $probe]

    if { $isDX } {
	if { [set pci [defget "PCI"]] == {} ||
	     [set jtag [defget "RESTRICTJTAGPINS"]] == {} } {
	    return {}
	}
	lappend cmd -pci $pci -jtag $jtag
    }
    return $cmd
}


# Export an auxilliary file like PIN, CRT or LOC and check if it is empty

proc exportAux { format file } {
    global argv0

    if { [catch "export -format $format $file" line] ||
	 [catch "open $file r" fp] } {
	puts stderr "$argv0: $line export $file failed"
	return {}
    }
    while { ([gets $fp line] >= 0) && [string match {;*} $line] } {
    }
    if { ([gets $fp line] < 0) || [string match -nocase {END.*} $line] } {
	close $fp
	catch "file delete -force -- $file"
	return {}
    }
    close $fp
    return $file
}


# Device name for display

proc deviceName { die pkg speed } {

    set name "${die}"
    if { ![string match "-*" $speed] } {
	append name "-"
    }
    return [append name "${speed}_${pkg}"]
}


# Query ADB and build migration commands

proc migrationCommands { _cmd newBase } {
    upvar 1 $_cmd cmd
    global argv0

    set placed [isPlaced]
    if { $placed < 0 } {
	return 2
    }

    set adlFile "${newBase}.adl"
    if { [catch "export -format adl $adlFile" msg] } {
	puts stderr "$argv0: $msg export netlist failed, aborting"
	return 3
    }

    set gotNewDie [mapDie $adlFile]
    if { $gotNewDie == {} } {
	return 4
    }
    foreach {origDie newFam newDie isDX isPlaceCompatible} $gotNewDie break

    set gotNewPkg [mapPackage $newDie]
    if { $gotNewPkg == {} } {
	return 5
    }
    foreach {origPkg newPkg} $gotNewPkg break

    set gotNewSpeed [mapSpeedGrade $newFam $newDie $newPkg]
    if { $gotNewSpeed == {} } {
	return 6
    }
    foreach {origSpeed newSpeed} $gotNewSpeed break

    set cmd(set_design) [setDesignCmd $newFam]
    if { $cmd(set_design) == {} } {
	return 7
    }

    set cmd(set_device) [setDeviceCmd $newFam $newDie $newPkg $newSpeed $isDX]
    if { $cmd(set_device) == {} } {
	return 8
    }

    set cmd(import_source) [list import_source -format "adl" $adlFile]

    set crtFile [exportAux "crt" "${newBase}.crt"]
    if { $crtFile != {} } {
	lappend cmd(import_source) -format "crt" $crtFile
    }

    if { $placed && $isPlaceCompatible } {
	set locFile [exportAux "loc" "${newBase}.loc"]
    } else {
	set locFile {}
    }

    if { $locFile == {} } {

	set cmd(import_loc) {}

	set pinFile [exportAux "pin" "${newBase}.pin"]
	if { $pinFile == {} } {
	    set cmd(pinStatus) " WARNING: Pin assignments did not exist in the original design."
	} else {
	    lappend cmd(import_source) -format "pin" $pinFile
	    set cmd(pinStatus) " Pin assignments transferred from the original design."
	}

    } else {

	set cmd(import_loc) [list import -format "loc" $locFile]
	set cmd(layout) [list layout -place_incremental "fix"]
	set cmd(pinStatus) ""

    }

    set cmd(compile) [list compile]

    set cmd(origDevice) [deviceName $origDie $origPkg $origSpeed]
    return 0
}


# Migrate ADB

proc migrate {} {
    global argv0

    set gotAdb [openAdb]
    if { $gotAdb == {} } {
	return 1
    }
    foreach {newAdb args} $gotAdb break

    array set cmd {}
    set status [migrationCommands cmd [file rootname $newAdb]]

    if { !$status } {
	if { [catch "eval $cmd(set_design)" msg] ||
	     [catch "eval $cmd(set_device)" msg] ||
	     [catch "eval $cmd(import_source)" msg] } {
	    set status 9
	    puts stderr "$argv0: $msg aborting"
	}
    }

    if { $status } {
	close_design
	set exts {"adl" "pin" "crt" "loc"}
	if { $args == 2 } {
	    lappend exts "adb"
	} else {
	    puts "Reverting back to the original design state saved in $newAdb"
	    catch "open_design $newAdb"
	}
	cleanup $newAdb $exts
	return $status
    }

    set newDevice [deviceName [defget DIE] [defget PACKAGE] [defget SPEED]]
    set designStatus " Please run Layout and verify timing."

    if { [catch "eval $cmd(compile)" msg] } {

	set designStatus "$msg WARNING: Compile errors need to be fixed."

    } elseif { $cmd(import_loc) != {} } {

	if { [catch "eval $cmd(import_loc)" msg] } {

	    set cmd(pinStatus) "$msg WARNING: Pin assignments need to be edited."

	} else {

	    set cmd(pinStatus) " Full placement transferred from the original design."

	    if { [catch "eval $cmd(layout)" msg] } {
		set designStatus "$msg WARNING: Layout is incomplete."
	    } else {
		set designStatus " Please verify timing and review the pin report."
	    }

	}
	cleanup $newAdb "loc"

    }

    if { $args == 2 } {
	save_design
    } else {
	puts "Original design state saved in $newAdb"
    }

    puts "\nMigrated design from $cmd(origDevice) to $newDevice"
    puts $cmd(pinStatus)
    puts $designStatus
}

migrate
